from abc import ABC
from typing import Tuple, Optional, Union, Callable

from qrogue.game.logic.base import StateVector, CircuitMatrix
from qrogue.game.logic.collectibles import Collectible
from qrogue.util import Logger


class Target(ABC):
    """
    Base class for fight-/puzzle-targets.
    """

    def __init__(self, id_: int, target: Union[StateVector, Callable[[], StateVector]], reward: Optional[Collectible],
                 input_: Optional[Union[StateVector, Callable[[], StateVector]]] = None,
                 allow_target_input_equality: bool = False):
        """
        Creates a Target with a given target state vector and a reward.
        :param id_: an integer unique per level to identify the target
        :param target: the StateVector to reach or a callable that returns it
        :param reward: the Collectible to get as a reward
        :param input_: the StateVector to start with (all |0> by default) or a callable that returns it
        :param allow_target_input_equality: whether target and input are allowed to be equal (would make basic puzzles
        be solved from the beginning, hence default is False)
        """
        self.__id = id_
        if isinstance(target, StateVector):
            self.__target: Callable[[], StateVector] = lambda: target
        else:
            self.__target: Callable[[], StateVector] = target

        if input_ is None:
            base_stv = StateVector.create_zero_state_vector(target.num_of_qubits)
            self.__input = lambda: base_stv
        elif isinstance(input_, StateVector):
            self.__input: Callable[[], StateVector] = lambda: input_
        else:
            self.__input: Callable[[], StateVector] = input_

        self.__reward: Optional[Collectible] = reward
        self.__is_active: bool = True
        # how often the target was checked (i.e., if it was reached) - can be used to determine the rewarded score
        self.__checks: int = 0

        if not allow_target_input_equality and self.__target().is_equal_to(self.__input(), ignore_god_mode=True):
            Logger.instance().warn(f"@Target.init(): target is equal to input (={self.__target})!", from_pycui=False)

    @property
    def id(self) -> int:
        return self.__id

    @property
    def state_vector(self) -> StateVector:
        """

        :return: this Target's StateVector
        """
        return self.__target()

    @property
    def input_stv(self) -> StateVector:
        return self.__input()

    @property
    def is_active(self) -> bool:
        """
        By default, Targets are active. They turn inactive as soon as their StateVector is reached.

        :return: whether this Target is still active or not
        """
        return self.__is_active

    @property
    def flee_energy(self) -> int:
        """

        :return: how much energy it costs to flee from this Target
        """
        return 1

    @property
    def checks(self) -> int:
        """
        Counts how often a player checked if the target StateVector was reached. Is used to determine the score gained
        from solving puzzles etc.

        :return: how often is_reached() was called
        """
        return self.__checks

    def is_reached(self, state_vector: StateVector, circ_matrix: CircuitMatrix) -> Tuple[bool, Optional[Collectible]]:
        """
        Checks if the given StateVector is equal to the Target's StateVector. If so, this Target is set inactive and
        will provide its reward.

        :param state_vector: the StateVector to check for equality
        :param circ_matrix: the CircuitMatrix used to create state_vector
        :return: True and a Collectible if the Target is reached, False and None otherwise
        """
        self.__checks += 1
        if self.__target().is_equal_to(state_vector):
            self._on_reached()
            self.__is_active = False
            temp = self.__reward
            self.__reward = None
            return True, temp
        return False, None

    def _on_reached(self):
        # Is called during a successful is_reached(). Gives Subclasses the possibility to react to the internal state
        # change.
        pass

    def __str__(self):
        string = "["
        for q in self.__target().to_value():
            string += f"{q} "
        string += "]"
        return string



import os

import qiskit  # imported to check its version

from qrogue.util.util_functions import cur_datetime
from .gameplay_config import CheatConfig
from .options import OptionsManager
from .path_config import PathConfig
from .test_config import TestConfig


class Config:  # todo make singleton and handle access to other configs?
    __frame_count = 0
    MAX_SEED = 1000000
    __VERSION = "v0.8.1"
    __GAME_CONFIG = "qrogue_game.config"
    __GAMEPLAY_HEAD = "[Gameplay]\n"
    __DEBUG = False
    __TEST_LEVEL = False

    __HEADER = "QRogue "
    __SEED_HEAD = "Seed="
    __TIME_HEAD = "Time="
    __CONFIG_HEAD = "[Config]"

    @staticmethod
    def ascii_art() -> str:
        return """


           _______           
          / _____ \          
         | |     | |         
         | |     | |         
         | |     | |         
         | |     | |         
         | |_____| |         
          \______\_\         

 
  _ __ ___   __ _ _   _  ___  
 | '__/ _ \ / _` | | | |/ _ \ 
 | | | (_) | (_| | |_| |  __/ 
 |_|  \___/ \__, |\__,_|\___| 
             __/ |            
            |___/             

"""

    @staticmethod
    def HEADER() -> str:
        return Config.__HEADER

    @staticmethod
    def SEED_HEAD() -> str:
        return Config.__SEED_HEAD

    @staticmethod
    def TIME_HEAD() -> str:
        return Config.__TIME_HEAD

    @staticmethod
    def CONFIG_HEAD() -> str:
        return Config.__CONFIG_HEAD

    @staticmethod
    def examiner_name() -> str:
        return "Examiner"  # for things the scientist says during the exam

    @staticmethod
    def scientist_name() -> str:
        return "Robb"

    @staticmethod
    def system_name() -> str:
        return "System"  # for things the robot's system tells us

    @staticmethod
    def get_name(value: int) -> str:
        if value == 0:
            return Config.examiner_name()
        if value == 1:
            return Config.scientist_name()
        if value == 2:
            return Config.system_name()

        return Config.system_name()  # default if nothing valid was specified

    @staticmethod
    def version() -> str:
        return Config.__VERSION

    @staticmethod
    def report_address() -> str:
        return "qrogue.game@gmail.com"

    @staticmethod
    def get_log_head(seed: int) -> str:
        now_str = cur_datetime().strftime("%d%m%Y_%H%M%S")
        head = f"{Config.__HEADER}{Config.version()}\n"
        head += f"{Config.__SEED_HEAD}{seed}\n"
        head += f"{Config.__TIME_HEAD}{now_str}\n\n"
        head += f"{Config.__CONFIG_HEAD}\n{OptionsManager.to_string()}\n"
        return head

    @staticmethod
    def game_config_file() -> str:
        return Config.__GAME_CONFIG

    @staticmethod
    def forbid_saving() -> bool:
        if Config.debugging():
            # in debugging saving is never forbidden
            return False
        return CheatConfig.did_cheat()

    @staticmethod
    def debugging() -> bool:
        return Config.__DEBUG

    @staticmethod
    def test_level(ignore_debugging: bool = False) -> bool:
        return Config.__TEST_LEVEL and (Config.debugging() or ignore_debugging)

    @staticmethod
    def debug_print(text: str):
        if Config.debugging():
            print(text)

    @staticmethod
    def activate_debugging(test_level: bool = False):
        Config.__DEBUG = True
        Config.__TEST_LEVEL = test_level

    @staticmethod
    def setup_user_data(custom_path: str = None):
        """
        Sets up the folder structure and default game config in the specified path.
        :param custom_path: where to store and load the user data (e.g. logs, save data)
        :return:
        """

        if custom_path:
            user_data_path = custom_path
        else:
            user_data_path = PathConfig.user_data_path()

        PathConfig.create_folder_structure(user_data_path)
        path = os.path.join(user_data_path, Config.__GAME_CONFIG)
        if not os.path.exists(path):
            text = ""
            text += Config.__GAMEPLAY_HEAD
            text += OptionsManager.to_string()
            with open(path, "x") as file:
                file.write(text)

    @staticmethod
    def load():
        if not os.path.exists(PathConfig.user_data_path(Config.__GAME_CONFIG)):
            Config.setup_user_data()

        if not Config.load_gameplay_config():
            return 2

        return 0

    @staticmethod
    def load_gameplay_config() -> bool:
        try:
            content = PathConfig.read(Config.game_config_file(), in_user_path=True)
            gameplay_start = content.index(Config.__GAMEPLAY_HEAD) + len(Config.__GAMEPLAY_HEAD)
            if OptionsManager.from_text(content[gameplay_start:len(content)]):
                return True
        except FileNotFoundError:
            raise FileNotFoundError("Could not load the game's config file!")
        return False

    @staticmethod
    def save_gameplay_config() -> bool:
        text = f"{Config.__GAMEPLAY_HEAD}\n{OptionsManager.to_string()}\n"
        PathConfig.write(Config.game_config_file(), text, in_user_path=True, may_exist=True, append=False)
        return True

    @staticmethod
    def skip_persisting() -> bool:
        return TestConfig.is_active()

    @staticmethod
    def inc_frame_count():
        Config.__frame_count += 1

    @staticmethod
    def frame_count() -> int:
        return Config.__frame_count

    @staticmethod
    def loading_refresh_time() -> float:
        return 0.2  # in seconds

    @staticmethod
    def using_new_qiskit() -> bool:
        parts = qiskit.__qiskit_version__['qiskit'].split(".")
        major = int(parts[0])
        minor = int(parts[1])
        return major >= 0 and minor >= 44  # >= 0.44.0

    @staticmethod
    def check_reachability(source: str, raise_exception: bool = False):
        print(f"\"{source}\" was reached.")
        if raise_exception:
            raise Exception(f"Faulty code at \"{source}\" was reached!")



# exporting
import qrogue.game.logic.actors
import qrogue.game.logic.base
import qrogue.game.logic.collectibles
from qrogue.game.logic.message import Message
from qrogue.game.logic.puzzle_generator import PuzzleGenerator

# importing
# +util



# exporting
from .collectible import Collectible, CollectibleType, MultiCollectible
from .collectible_factory import CollectibleFactory, GateFactory, OrderedCollectibleFactory
from .instruction import Instruction, InstructionManager
from .pickup import Pickup, Score, Key, Energy, Qubit
from .special_items import QuantumFuser

# importing
# +base
# +util



from typing import Callable, Tuple

import py_cui
from py_cui.errors import PyCUIOutOfBoundsError


class PyCuiConfig:
    # 0 is used as default since it isn't a valid width or height otherwise
    __get_dimensions: Callable[[], Tuple[int, int]] = lambda: (0, 0)

    @staticmethod
    def get_min_dimensions() -> Tuple[int, int]:
        """

        Returns: minimum number of rows and columns needed to display the game correctly

        """
        # retrieved by trial and error
        # 133 is the minimum width to fully display a 2-qubit puzzle without visual errors
        # 31 can be explained like this: Our PyCUI has 10 rows, so there can be 10 widgets per column. To display a
        # widget correctly, PyCUI needs at least three character-rows - 1x upper border, 1x content, 1x lower border.
        # Hence, one PyCUI row corresponds to at least three character-rows. Therefore, we need at least 30 character-
        # rows to display all widgets. Furthermore, one row is needed to display the title bar. The result is a minimum
        # of 31 rows. We're not sure whether that is exactly how PyCUI works under the hood, but it makes sense and
        # everything below 31 rows throws an internal error complaining about the height being too small.
        return 31, 133

    @staticmethod
    def set_get_dimensions_callback(get_dimensions: Callable[[], Tuple[int, int]]):
        PyCuiConfig.__get_dimensions = get_dimensions

    @staticmethod
    def get_dimensions() -> Tuple[int, int]:
        return PyCuiConfig.__get_dimensions()

    KEY_ESCAPE = py_cui.keys.KEY_ESCAPE
    KEY_CTRL_Q = py_cui.keys.KEY_CTRL_Q

    OutOfBoundsError = PyCUIOutOfBoundsError
    PyCuiWidget = py_cui.widgets.Widget


class PyCuiColors:
    BLACK_ON_WHITE = py_cui.BLACK_ON_WHITE
    BLACK_ON_RED = py_cui.BLACK_ON_RED

    BLUE_ON_BLACK = py_cui.BLUE_ON_BLACK
    BLUE_ON_WHITE = py_cui.BLUE_ON_WHITE

    CYAN_ON_BLACK = py_cui.CYAN_ON_BLACK

    GREEN_ON_BLACK = py_cui.GREEN_ON_BLACK
    GREEN_ON_WHITE = py_cui.GREEN_ON_WHITE

    MAGENTA_ON_BLACK = py_cui.MAGENTA_ON_BLACK
    MAGENTA_ON_WHITE = py_cui.MAGENTA_ON_WHITE

    RED_ON_BLACK = py_cui.RED_ON_BLACK
    RED_ON_BLUE = py_cui.RED_ON_BLUE
    RED_ON_WHITE = py_cui.RED_ON_WHITE

    WHITE_ON_BLACK = py_cui.WHITE_ON_BLACK
    WHITE_ON_CYAN = py_cui.WHITE_ON_CYAN
    WHITE_ON_GREEN = py_cui.WHITE_ON_GREEN
    WHITE_ON_MAGENTA = py_cui.WHITE_ON_MAGENTA

    YELLOW_ON_BLACK = py_cui.YELLOW_ON_BLACK



import math
from abc import ABC, abstractmethod
from typing import Iterator, Optional, Dict, List, Callable, Tuple, Any

import qiskit.circuit.library.standard_gates as gates
from qiskit.circuit import Gate as QiskitGate

from qrogue.game.logic.base import StateVector, CircuitMatrix, QuantumSimulator, QuantumCircuit, UnitarySimulator
from qrogue.game.logic.collectibles import Collectible, CollectibleType
from qrogue.util import Logger, GateType, QuantumSimulationConfig, InstructionConfig, ColorConfig, ColorCode, \
    SaveGrammarConfig
from qrogue.util.achievements import Unlocks
from qrogue.util.util_functions import rad2deg, center_string, to_binary_string, num_to_letter


class Instruction(Collectible, ABC):
    """
    Wrapper class for gates from qiskit.circuit.library with their needed arguments (qubits/cbits to apply it on)
    """
    MAX_ABBREVIATION_LEN = 5

    @staticmethod
    def compute_stv(instructions: List["Instruction"], num_of_qubits: int, inverse: bool = False) -> StateVector:
        circuit = QuantumCircuit.from_bit_num(num_of_qubits, num_of_qubits)
        for instruction in instructions:
            instruction.append_to(circuit, inverse)
        simulator = QuantumSimulator()
        amplitudes = simulator.run(circuit, do_transpile=True)
        return StateVector(amplitudes, num_of_used_gates=len(instructions))

    @staticmethod
    def __circuit_input_value(qubit: int, state_vectors: Optional[Tuple[StateVector, StateVector, StateVector]]):
        if state_vectors is not None:
            input_stv, output_stv, target_stv = state_vectors
            if input_stv is not None and input_stv.is_classical \
                    and target_stv is not None and target_stv.is_classical \
                    and output_stv.is_classical:  # robot.state_vector cannot be None
                index = input_stv.to_value().index(1)  # find where the amplitude is 1
                # get the respective qubit values but in lsb, so we can use $qubit directly as index
                values = to_binary_string(index, input_stv.num_of_qubits, msb=False)
                return f"= {values[qubit]} "
        return ""

    @staticmethod
    def __circuit_output_value(qubit: int, state_vectors: Optional[Tuple[StateVector, StateVector, StateVector]]) -> str:
        if state_vectors is not None:
            input_stv, output_stv, target_stv = state_vectors
            if input_stv is not None and input_stv.is_classical \
                    and target_stv is not None and target_stv.is_classical \
                    and output_stv.is_classical:  # robot.state_vector cannot be None
                index = output_stv.to_value().index(1)  # find where the amplitude is 1
                # get the respective qubit values but in lsb, so we can use $qubit directly as index
                out_values = to_binary_string(index, output_stv.num_of_qubits, msb=False)
                index = target_stv.to_value().index(1)
                target_values = to_binary_string(index, target_stv.num_of_qubits, msb=False)
                is_correct = out_values[qubit] == target_values[qubit]
                equality = ColorConfig.colorize(ColorCode.PUZZLE_CORRECT_AMPLITUDE if is_correct
                                                else ColorCode.PUZZLE_WRONG_AMPLITUDE,
                                                '=' + ('=' if is_correct else '/') + '=')
                return f"= {out_values[qubit]}| {equality} <{target_values[qubit]}|"
        return "|"

    @staticmethod
    def circuit_to_string(num_of_qubits: int, circuit_space: int, instructions: Dict[int, "Instruction"],
                          preview: Optional[Tuple[Optional["Instruction"], int, int]] = None,
                          state_vectors: Optional[Tuple[StateVector, StateVector, StateVector]] = None) -> str:
        """
        :param num_of_qubits: how many rows (i.e., qubits) the circuit has
        :param circuit_space: how many columns (i.e., places for instructions)
        :param instructions: all positions within the circuit where an Instruction is placed
        :param preview: optionally describes a gate that is not yet placed onto the circuit
        :param state_vectors: a tuple consisting of an Input-, Output- and Target-StateVector to potentially show a
            qubit's (classical) value
        :return: a string representing the described circuit
        """
        entry = "-" * (3 + InstructionConfig.MAX_ABBREVIATION_LEN + 3)
        rows = [[entry] * circuit_space for _ in range(num_of_qubits)]
        for i, inst in instructions.items():
            for q in inst.qargs_iter():
                inst_str = center_string(inst.abbreviation(q), InstructionConfig.MAX_ABBREVIATION_LEN)
                rows[q][i] = f"--{{{inst_str}}}--"

        if preview is not None:
            inst, pos, qubit = preview
            if inst is None:
                rows[qubit][pos] = "--/   /--"
            else:
                for q in inst.qargs_iter():
                    rows[q][pos] = f"--{{{inst.abbreviation(q)}}}--"
                rows[qubit][pos] = f"-- {inst.abbreviation(qubit)} --"

        # every line consists of ket-pre- and -suffix with the entries (empty or instruction) separated by +
        lines = [f"| q{q} {Instruction.__circuit_input_value(q, state_vectors)}>" + "+".join(rows[q])
                 + f"< q'{q} {Instruction.__circuit_output_value(q, state_vectors)}" for q in range(num_of_qubits)]
        lines.reverse()  # place qubits from top to bottom, high to low index

        # for some reason the whitespace in front is needed to center the text correctly
        lines[0] = "In " + lines[0] + " Out"
        line_width = len(lines[0])
        return "\n".join([center_string(line, line_width, uneven_left=False) for line in lines])

    def __init__(self, gate_type: GateType, instruction: QiskitGate, needed_qubits: int):
        super().__init__(CollectibleType.Gate)
        self.__type = gate_type
        self.__instruction = instruction
        self.__needed_qubits = needed_qubits
        self._qargs = []
        self._cargs = []
        self.__position: Optional[int] = None

    @property
    def gate_type(self) -> GateType:
        return self.__type

    @property
    def num_of_qubits(self) -> int:
        return self.__needed_qubits

    @property
    def position(self) -> Optional[int]:
        return self.__position

    @property
    def no_qubits_specified(self) -> bool:
        return len(self._qargs) <= 0

    @property
    def difficulty(self) -> int:
        """
        Describes how difficulty it is to use this Instruction (higher number = higher difficulty).

        :return: positive integer
        """
        return self.__type.difficulty

    def can_use_qubit(self, qubit: int) -> bool:
        return qubit not in self._qargs

    def use_qubit(self, qubit: int) -> bool:
        """

        :param qubit: the qubit to use for this Instruction
        :return: True if more qubits are needed for the Instruction to work, False if there are enough
        """
        if len(self._qargs) >= self.__needed_qubits:
            return False
        if not self.can_use_qubit(qubit):
            Logger.instance().error("Cannot use the same qubit multiple times!", from_pycui=False)
            return True
        self._qargs.append(qubit)
        return len(self._qargs) < self.__needed_qubits

    def is_used(self) -> bool:
        return self.__position is not None

    def use(self, position: int) -> bool:
        """
        Checks if this Instruction has all its needed information (e.g. qubit arguments) specified and then saves the
        given position.
        :param position: position where we want to use this Instruction at
        :return: True if this Instruction was successfully used/positioned, False otherwise
        """
        assert position >= 0

        if len(self._qargs) == self.__needed_qubits:
            self.__position = position
            return True
        return False

    def setup(self, qargs: List[int], cargs: Optional[List[int]] = None, position: Optional[int] = None) \
            -> "Instruction":
        """
        Function used to programmatically set up a gate in one go after instantiation.
        """
        if len(qargs) > self.__needed_qubits:
            Logger.instance().warn(f"Instruction.setup(): {len(qargs)} qubits provided but only {self.__needed_qubits} "
                                   f"needed. Ignoring the over-specified ones.", from_pycui=False)
        if len(qargs) < self.__needed_qubits:
            Logger.instance().warn(f"Instruction.setup(): {len(qargs)} qubits provided but {self.__needed_qubits} "
                                   f"needed. Things might not work correctly due to under-specified qubits.",
                                   from_pycui=False)
        for qubit in qargs:
            self.use_qubit(qubit)
        if cargs is not None:
            self._cargs = cargs
        if position is not None:
            self.use(position)
        return self

    def reset(self, skip_qargs: bool = False, skip_position: bool = False):
        if not skip_qargs:
            self._qargs = []
        if not skip_position:
            self.__position = None

    def append_to(self, circuit: QuantumCircuit, inverse: bool = False):
        instruction = self.__instruction.inverse() if inverse else self.__instruction
        circuit.append(instruction, self._qargs, self._cargs)

    def qargs_iter(self) -> Iterator[int]:
        return iter(self._qargs)

    def get_qubit_at(self, index: int = 0) -> int:
        if 0 <= index < len(self._qargs):
            return self._qargs[index]

    def name(self, include_suffix: Optional[bool] = None) -> str:
        if include_suffix is None: include_suffix = True
        return f"{self.__type.short_name}{' Gate' if include_suffix else ''}"

    def description(self, check_unlocks: Optional[Callable[[str], bool]] = None) -> str:
        desc = f"Full name: {self.gate_type.full_name}\n"
        desc += f"Abbreviation: {self.abbreviation(qubit=None).strip()}\n"
        desc += self.__type.description
        if check_unlocks is not None and check_unlocks(Unlocks.ShowEquation.ach_name):
            desc += "\n\nMatrix:\n" + self._matrix_string()
        return desc

    @abstractmethod
    def abbreviation(self, qubit: Optional[int] = None) -> str:
        pass

    def _matrix_string(self) -> str:
        # use the real underlying matrix because some Instructions might have parameters
        return CircuitMatrix.matrix_to_string(self.__instruction.to_matrix(), self.num_of_qubits)

    @abstractmethod
    def copy(self) -> "Instruction":
        pass

    def _deep_copy(self) -> "Instruction":
        return self.copy().setup(self._qargs, self._cargs, self.__position)

    def selection_str(self) -> str:
        # Gate (qX, qY, ?, ...)
        text = f"{self.name()} ("
        for i in range(self.num_of_qubits - 1):
            if i < len(self._qargs):
                text += f"q{self._qargs[i]}, "
            else:
                text += "?, "
        if self.num_of_qubits - 1 < len(self._qargs):
            text += f"q{self._qargs[self.num_of_qubits - 1]})"
        else:
            text += "?)"
        if self.is_used():
            text += f" @{self.__position}"
        return text

    def preview_str(self, next_qubit: int) -> str:
        # Gate (qX, qY, ?, ...)
        self._qargs.append(next_qubit)  # pretend that we already set the next qubit
        preview = self.selection_str()
        self._qargs.pop()  # undo setting the next qubit because we only wanted to pretend that we did
        return preview

    def to_save_string(self) -> str:
        return self.__type.short_name

    def to_string(self):
        return self.name()

    def __str__(self) -> str:
        text = self.to_string()
        if self.__position is not None: text += f" [pos={self.__position}]"
        if len(self._qargs) > 0: text += f" @q{self._qargs[0]}"
        for i in range(1, len(self._qargs)): text += f", {self._qargs[i]}"
        return text


####### Single Qubit Gates #######


class SingleQubitGate(Instruction, ABC):
    def __init__(self, gate_type: GateType, instruction: QiskitGate):
        super().__init__(gate_type, instruction, needed_qubits=1)


class IGate(SingleQubitGate):
    def __init__(self):
        super().__init__(GateType.IGate, gates.IGate())

    def abbreviation(self, qubit: Optional[int] = None):
        return "I"

    def copy(self) -> "Instruction":
        return IGate()


class XGate(SingleQubitGate):
    def __init__(self):
        super(XGate, self).__init__(GateType.XGate, gates.XGate())

    def abbreviation(self, qubit: Optional[int] = None):
        return " X "

    def copy(self) -> "Instruction":
        return XGate()


class SXGate(SingleQubitGate):
    def __init__(self):
        super(SXGate, self).__init__(GateType.SXGate, gates.SXGate())

    def abbreviation(self, qubit: Optional[int] = None):
        return "SX "

    def copy(self) -> "Instruction":
        return SXGate()


class YGate(SingleQubitGate):
    def __init__(self):
        super(YGate, self).__init__(GateType.YGate, gates.YGate())

    def abbreviation(self, qubit: Optional[int] = None):
        return " Y "

    def copy(self) -> "Instruction":
        return YGate()


class ZGate(SingleQubitGate):
    def __init__(self):
        super(ZGate, self).__init__(GateType.ZGate, gates.ZGate())

    def abbreviation(self, qubit: Optional[int] = None):
        return " Z "

    def copy(self) -> "Instruction":
        return ZGate()


class HGate(SingleQubitGate):
    def __init__(self):
        super().__init__(GateType.HGate, gates.HGate())

    def abbreviation(self, qubit: Optional[int] = None):
        return " H "

    def copy(self) -> "Instruction":
        return HGate()


class SGate(SingleQubitGate):
    def __init__(self):
        super().__init__(GateType.SGate, gates.SGate())

    def abbreviation(self, qubit: Optional[int] = None):
        return " S "

    def copy(self) -> "Instruction":
        return SGate()


class RotationGate(SingleQubitGate, ABC):
    _DEFAULT_ANGLE = math.pi / 2

    def __init__(self, gate_type: GateType, instruction: QiskitGate, angle: float):
        """
        :param angle: value between 0 and PI (other values possible, but they correspond to an angle in said range)
        """
        super().__init__(gate_type, instruction)
        self.__angle = angle

    @property
    def angle(self) -> float:
        return self.__angle

    def description(self, check_unlocks: Optional[Callable[[str], bool]] = None) -> str:
        desc = super().description(check_unlocks)  # remove the stated default angle at the end
        # find indices of "°" and the whitespace before that, so we can replace the angle value.
        degree_index = desc.find("°")
        if degree_index <= 0:
            Logger.instance().error(f"No \"°\" in the description of {self.type}!", show=False, from_pycui=False)
            return desc
        space_index = desc[:degree_index].rfind(" ")

        # round and skip potential ".00"s
        angle = str(round(rad2deg(self.angle), 2))
        if angle.endswith(".00"): angle = angle[:-3]

        return f"{desc[:space_index]} {angle}{desc[degree_index:]}"


class RYGate(RotationGate):
    def __init__(self, angle: float = RotationGate._DEFAULT_ANGLE):
        super().__init__(GateType.RYGate, gates.RYGate(theta=angle), angle)

    def abbreviation(self, qubit: Optional[int] = None):
        return " RY"

    def copy(self) -> "Instruction":
        return RYGate(self.angle)


class RZGate(RotationGate):
    def __init__(self, angle: float = RotationGate._DEFAULT_ANGLE):
        super().__init__(GateType.RZGate, gates.RZGate(phi=angle), angle)

    def abbreviation(self, qubit: Optional[int] = None):
        return " RZ"

    def copy(self) -> "Instruction":
        return RZGate(self.angle)


####### Multi Qubit Gates ########


class MultiQubitGate(Instruction, ABC):
    def abbreviation(self, qubit: Optional[int] = None):
        if qubit is None:
            index = None
        elif qubit in self._qargs:
            index = self._qargs.index(qubit)
        else:
            # during placement, we need an abbreviation for the next qubit
            index = len(self._qargs)
        return self._internal_abbreviation(index)

    @abstractmethod
    def _internal_abbreviation(self, index: Optional[int]) -> str:
        pass


####### Double Qubit Gates #######


class DoubleQubitGate(MultiQubitGate, ABC):
    def __init__(self, gate_type: GateType, instruction: QiskitGate):
        super(DoubleQubitGate, self).__init__(gate_type, instruction, needed_qubits=2)


class SwapGate(DoubleQubitGate):
    def __init__(self):
        super().__init__(GateType.SwapGate, gates.SwapGate())

    def _internal_abbreviation(self, index: Optional[int]) -> str:
        if index is None:
            return "SW "
        else:
            return f"SW{index}"

    def copy(self) -> "Instruction":
        return SwapGate()


class CXGate(DoubleQubitGate):
    def __init__(self):
        super().__init__(GateType.CXGate, gates.CXGate())

    def _internal_abbreviation(self, index: Optional[int]) -> str:
        if index is None:
            return "C/X"
        elif index == 0:
            return " C "
        elif index == 1:
            return " X "

    def copy(self) -> "Instruction":
        return CXGate()


class CYGate(DoubleQubitGate):
    def __init__(self):
        super().__init__(GateType.CYGate, gates.CYGate())

    def _internal_abbreviation(self, index: Optional[int]) -> str:
        if index is None:
            return "C/Y"
        elif index == 0:
            return " C "
        elif index == 1:
            return " Y "

    def copy(self) -> "Instruction":
        return CYGate()


class CZGate(DoubleQubitGate):
    def __init__(self):
        super().__init__(GateType.CZGate, gates.CZGate())

    def _internal_abbreviation(self, index: Optional[int]) -> str:
        if index is None:
            return "C/Z"
        elif index == 0:
            return " C "
        elif index == 1:
            return " Z "

    def copy(self) -> "Instruction":
        return CZGate()


class CHGate(DoubleQubitGate):
    def __init__(self):
        super().__init__(GateType.CHGate, gates.CHGate())

    def _internal_abbreviation(self, index: Optional[int]) -> str:
        if index is None:
            return "C/H"
        elif index == 0:
            return " C "
        elif index == 1:
            return " H "

    def copy(self) -> "Instruction":
        return CHGate()


####### Combined Gates #######

class CombinedGate(MultiQubitGate):
    __NEXT_ID = 0

    @staticmethod
    def validate_instructions(instructions: List[Instruction], num_quantum_fusers: Optional[int] = None) -> \
            Tuple[int, Any]:
        """
        Exit codes:
            - 0 = valid
            - 1 = not enough Instructions, returns number of given instructions
            - 2 = too many Instructions, returns number of given instructions
            - 3 = not enough QuantumFusers, returns number of needed QuantumFusers
            - 4 = contains a CombinedGate, returns name of CombinedGate
            - 5 = an Instruction is not set up, returns name of Instruction

        :param instructions: list of Instructions used to build a CombinedGate
        :param num_quantum_fusers: optional parameter to check whether we have enough quantum fusers
        :return: (0, None) if instructions are valid, other values depending on which part was invalid
        """
        if len(instructions) < InstructionConfig.COMB_GATE_MIN_GATE_NUM: return 1, len(instructions)
        if len(instructions) > InstructionConfig.COMB_GATE_MAX_GATE_NUM: return 2, len(instructions)
        if num_quantum_fusers is not None and len(instructions) > num_quantum_fusers: return 3, len(instructions)
        for inst in instructions:
            if inst.gate_type is GateType.Combined:
                return 4, inst.name()
            if inst.no_qubits_specified:
                return 5, inst.name()
        return 0, None

    @staticmethod
    def instructions_criteria() -> str:
        return "A CombinedGate is not allowed to be built from another CombinedGate."

    @staticmethod
    def validate_gate_name(name: str) -> int:
        """
        Exit codes:
            - 0 = valid
            - 1 = not enough characters
            - 2 = too many characters
            - 3 = contains an illegal (i.e., non-letter) character
            - 4 = equal to the name of a base gate

        :param name: the name to validate
        :return: 0 if name is valid, other values depending on which part was invalid
        """
        # check for correct number of letters
        if len(name) < InstructionConfig.COMB_GATE_NAME_MIN_CHARACTERS: return 1
        if len(name) > InstructionConfig.COMB_GATE_NAME_MAX_CHARACTERS: return 2
        # check if name only consists of letters
        if not name.isalpha(): return 3
        if name in InstructionManager.gate_names(include_suffix=False): return 4
        # all criteria fulfilled
        return 0

    @staticmethod
    def gate_name_criteria() -> str:
        return f"The name needs to consist of {InstructionConfig.COMB_GATE_NAME_MIN_CHARACTERS} to " \
               f"{InstructionConfig.COMB_GATE_NAME_MAX_CHARACTERS} letters (no numbers or other characters allowed). " \
               f"You don't have to add a \"Gate\"-suffix, the game will do this automatically where needed."

    @staticmethod
    def _next_id() -> int:
        next_id = CombinedGate.__NEXT_ID
        CombinedGate.__NEXT_ID += 1
        return next_id

    def __init__(self, instructions: List[Instruction], needed_qubits: int, name: Optional[str] = None,
                 _id: Optional[int] = None):
        inst_validation = CombinedGate.validate_instructions(instructions)
        name_validation = CombinedGate.validate_gate_name(name)
        Logger.instance().assertion(inst_validation[0] == 0, f"Instructions are not valid: "
                                                             f"exit code={inst_validation}, data={inst_validation[1]}")
        Logger.instance().assertion(name_validation == 0, f"Name \"{name}\" is not valid: exit code {name_validation}")

        if name is None: name = "BlackBox"
        if name.endswith("Gate"): name = name[:-len("Gate")]
        name = name.strip()   # remove leftover whitespaces

        self.__id = CombinedGate._next_id() if _id is None else _id
        # calculate difficulty based on the most difficult instruction used plus how many other instructions are used
        self.__difficulty = max([inst.difficulty for inst in instructions]) + len(instructions) - 1

        # order instructions based on their position property or position in the list if the property is None
        ##################################################
        cur_pos = 0
        inst_dict: Dict[int, List[Instruction]] = {}
        for i, inst in enumerate(instructions):
            if inst.position is not None:
                cur_pos = inst.position     # update current position
            # else: inst has no valid position, so we place it directly after the last valid/current position
            # append inst
            if cur_pos in inst_dict:
                inst_dict[cur_pos].append(inst)
            else:
                inst_dict[cur_pos] = [inst]

        instructions = []
        sorted_keys = list(inst_dict.keys())
        sorted_keys.sort()
        for pos in sorted_keys:
            instructions += inst_dict[pos]
        ##################################################

        circuit = QuantumCircuit.from_register(needed_qubits)
        for inst in instructions: inst.append_to(circuit)
        gate = circuit.to_gate(label=name)

        super().__init__(GateType.Combined, gate, needed_qubits)
        self.__gate = gate
        self.__instructions = instructions

        inst_dict: Dict[int, Instruction] = {}
        for i, inst in enumerate(instructions): inst_dict[i] = inst
        self.__circ_repr = Instruction.circuit_to_string(needed_qubits, len(instructions), inst_dict)

        amplitudes = UnitarySimulator().execute(circuit, decimals=QuantumSimulationConfig.DECIMALS)
        self.__matrix = CircuitMatrix(amplitudes, len(instructions))

    @property
    def difficulty(self) -> int:
        return self.__difficulty

    def name(self, include_suffix: Optional[bool] = None) -> str:
        if include_suffix is None: include_suffix = True
        return f"{self.__gate.label}{' Gate' if include_suffix else ''}"

    def _internal_abbreviation(self, index: Optional[int] = None):
        index = "" if index is None else str(index)
        id_str = str(self.__id) if self.__id < 10 else num_to_letter(self.__id - 10, start_uppercase=True)
        return f"{GateType.Combined.short_name[0]}{id_str}{index}"

    def _matrix_string(self) -> str:
        return self.__matrix.to_string()

    def description(self, check_unlocks: Optional[Callable[[str], bool]] = None) -> str:
        desc = super().description(check_unlocks)
        # remove the first line ("Full name:") because it's meaningless for CombinedGates with custom names
        desc = desc[desc.index("\n")+1:]
        desc += "\nUnderlying Circuit:\n"
        desc += self.__circ_repr
        return desc

    def to_save_string(self) -> str:
        # e.g.: combined{"Bell" 2: H(0),CX(0,1)}
        text = [f"{inst.to_save_string()}({','.join([str(q) for q in inst.qargs_iter()])})"
                for inst in self.__instructions]
        return f"{SaveGrammarConfig.combined_prefix()}{{\"{self.__gate.label}\" {self.num_of_qubits}: " \
               f"{','.join(text)}}}"

    def copy(self) -> "Instruction":
        # copy the instructions to not accidentally alter the positioning of the underlying Instructions
        instructions = [inst._deep_copy() for inst in self.__instructions]
        return CombinedGate(instructions, self.__gate.num_qubits, self.__gate.label, self.__id)


####### Gates for internal use only #######

class DebugGate(SingleQubitGate):
    def __init__(self):
        super().__init__(GateType.Debug, gates.RZGate(phi=2.5))

    def abbreviation(self, qubit: Optional[int] = None):
        return "deb"

    def copy(self) -> "Instruction":
        return DebugGate()


class InstructionManager:
    __GATES: Dict[GateType, Instruction] = {
        GateType.IGate: IGate(),

        GateType.XGate: XGate(),
        GateType.SXGate: SGate(),
        GateType.YGate: YGate(),
        GateType.ZGate: ZGate(),

        GateType.HGate: HGate(),
        GateType.SGate: SGate(),

        GateType.RYGate: RYGate(),
        GateType.RZGate: RZGate(),

        GateType.SwapGate: SwapGate(),
        GateType.CXGate: CXGate(),
        GateType.CYGate: CYGate(),
        GateType.CZGate: CZGate(),
        GateType.CHGate: CHGate(),

        GateType.Debug: DebugGate(),
    }

    @staticmethod
    def validate() -> bool:
        for val in GateType:
            if val is GateType.Combined: continue  # todo: properly implement for Combined? Or do we have to skip them?
            assert val in InstructionManager.__GATES, f"{val} not defined in InstructionManager.__GATES!"
        return True

    @staticmethod
    def gate_names(include_suffix: Optional[bool] = None) -> List[str]:
        return [gate.name(include_suffix) for gate in InstructionManager.__GATES.values()]

    @staticmethod
    def from_type(gate_type: GateType) -> Instruction:
        return InstructionManager.__GATES[gate_type].copy()

    @staticmethod
    def type_from_name(name: str, ignore_gate_suffix: bool = True) -> Optional[GateType]:
        if ignore_gate_suffix and name.lower().endswith("gate"):
            name = name[:-len("gate")]
        for gate_type in GateType:
            if gate_type.is_in_names(name):
                return gate_type
        return None

    @staticmethod
    def instruction_from_name(name: str, ignore_gate_suffix: bool = True) -> Optional[Instruction]:
        gate_type = InstructionManager.type_from_name(name, ignore_gate_suffix)
        if gate_type is None:
            return None
        return InstructionManager.from_type(gate_type)



class TestConfig:
    class StateException(Exception):
        def __init__(self, msg: str):
            super(TestConfig.StateException, self).__init__(msg)
            self.__msg = msg

        def __str__(self):
            return f"Qrogue TestStateException: {self.__msg}"

    __IS_ACTIVE = False
    __IS_AUTOMATIC = True

    @staticmethod
    def is_active() -> bool:
        return TestConfig.__IS_ACTIVE

    @staticmethod
    def is_automatic() -> bool:
        return TestConfig.__IS_AUTOMATIC

    @staticmethod
    def automation_step_time() -> int:
        """
        For automated tests the test simulation runs automatically without user intervention.

        :return: how long to wait between each step in ms
        """
        return 10

    @staticmethod
    def key_pause() -> float:
        return 0.05  # time in seconds

    @staticmethod
    def activate():
        TestConfig.__IS_ACTIVE = True

    @staticmethod
    def deactivate():
        TestConfig.__IS_ACTIVE = False

    @staticmethod
    def set_automation(automatic: bool = True):
        TestConfig.__IS_AUTOMATIC = automatic

    @staticmethod
    def test_seed() -> int:
        if TestConfig.is_active():
            return 7
        else:
            raise Exception("TestConfig not yet activated!")



from qrogue.game.logic.actors.controllables import ControllableType
from qrogue.game.world.tiles import TileCode
from qrogue.util import PyCuiColors, ColorConfig
from ..widget_base import WidgetWrapper


class _TileColorer:  # todo: move to ColorConfig? (only value of dictionary since the keys are TileCodes)
    __color_manager = {
        TileCode.Invalid: PyCuiColors.RED_ON_BLUE,
        TileCode.Void: PyCuiColors.CYAN_ON_BLACK,
        TileCode.FogOfWar: PyCuiColors.CYAN_ON_BLACK,

        TileCode.Floor: PyCuiColors.CYAN_ON_BLACK,
        TileCode.Wall: PyCuiColors.BLACK_ON_WHITE,
        TileCode.Obstacle: PyCuiColors.BLACK_ON_WHITE,
        TileCode.Door: PyCuiColors.CYAN_ON_BLACK,

        TileCode.Collectible: PyCuiColors.CYAN_ON_BLACK,
        TileCode.Teleport: PyCuiColors.CYAN_ON_BLACK,
        TileCode.Message: PyCuiColors.CYAN_ON_BLACK,
        TileCode.Goal: PyCuiColors.GREEN_ON_BLACK,

        TileCode.Controllable: PyCuiColors.WHITE_ON_GREEN,
        TileCode.Enemy: PyCuiColors.RED_ON_BLACK,
        TileCode.Boss: PyCuiColors.BLACK_ON_RED,
    }

    @staticmethod
    def get_color(tile_code: TileCode) -> int:
        """

        :param tile_code: code of the Tile we want to get the default color of
        :return: integer representing one of the possible foreground-background color comibnations, None for invalid
        input
        """
        if tile_code in _TileColorer.__color_manager:
            return _TileColorer.__color_manager[tile_code]


class ColorRules:
    @staticmethod
    def apply_map_rules(widget: WidgetWrapper) -> None:
        # controllables
        for ct in ControllableType.values():
            widget.add_text_color_rule(ct.name, _TileColorer.get_color(TileCode.Controllable), 'contains',
                                       match_type='regex')
        # puzzles
        widget.add_text_color_rule(TileCode.Boss.representation, _TileColorer.get_color(TileCode.Boss), 'contains',
                                   match_type='regex')
        widget.add_text_color_rule('\d', _TileColorer.get_color(TileCode.Enemy), 'contains', match_type='regex')
        widget.add_text_color_rule(f"\{TileCode.Riddler.representation}", _TileColorer.get_color(TileCode.Enemy),
                                   'contains', match_type='regex')
        widget.add_text_color_rule(TileCode.Challenger.representation, _TileColorer.get_color(TileCode.Enemy),
                                   'contains', match_type='regex')

        # blocking
        widget.add_text_color_rule(TileCode.Wall.representation, _TileColorer.get_color(TileCode.Wall), 'contains',
                                   match_type='regex')
        widget.add_text_color_rule(TileCode.Obstacle.representation, _TileColorer.get_color(TileCode.Obstacle),
                                   'contains', match_type='regex')

        # collectibles
        for ct in TileCode.collectible_subtypes():
            widget.add_text_color_rule(ct.representation, _TileColorer.get_color(TileCode.Collectible), 'contains',
                                       match_type='regex')

        # goal
        widget.add_text_color_rule(TileCode.Goal.representation, _TileColorer.get_color(TileCode.Goal),
                                   'contains', match_type='regex')

    @staticmethod
    def apply_navigation_rules(widget: WidgetWrapper):
        widget.add_text_color_rule('#', _TileColorer.get_color(TileCode.Wall), 'contains', match_type='regex')
        widget.add_text_color_rule('M', _TileColorer.get_color(TileCode.Controllable), 'contains', match_type='regex')
        widget.add_text_color_rule('t', _TileColorer.get_color(TileCode.Teleport), 'contains', match_type='regex')
        widget.add_text_color_rule('\.', _TileColorer.get_color(TileCode.Message), 'contains', match_type='regex')

    @staticmethod
    def apply_circuit_rules(widget: WidgetWrapper):
        # highlight everything between {} (gates), |> (start) or <| (end) or | | (In/Out label)
        regex_gates = "\{.*?\}"
        regex_start = "\|.*?\>"
        regex_end = "\<.*?\|"
        widget.add_text_color_rule(f"({regex_gates}|{regex_start}|{regex_end})", ColorConfig.CIRCUIT_COLOR, 'contains',
                                   match_type='regex')
        regex_label = "In|Out"
        widget.add_text_color_rule(f"({regex_label})", ColorConfig.CIRCUIT_LABEL_COLOR, 'contains', match_type='regex')

    @staticmethod
    def apply_heading_rules(widget: WidgetWrapper):
        widget.add_text_color_rule("~.*~", ColorConfig.STV_HEADING_COLOR, 'contains', match_type='regex')

    @staticmethod
    def apply_qubit_config_rules(widget: WidgetWrapper):
        widget.add_text_color_rule("\|.*>", ColorConfig.QUBIT_CONFIG_COLOR, 'contains', match_type='regex')

    @staticmethod
    def apply_level_selection_seed_rules(widget: WidgetWrapper):
        widget.add_text_color_rule("Seed:", ColorConfig.LEVEL_SELECTION_HEADER_COLOR, 'contains', match_type='regex')

    @staticmethod
    def apply_level_selection_level_rules(widget: WidgetWrapper):
        widget.add_text_color_rule("Level:", ColorConfig.LEVEL_SELECTION_HEADER_COLOR, 'contains', match_type='regex')

    @staticmethod
    def apply_transition_rules(widget: WidgetWrapper):
        widget.add_text_color_rule("\[.*\]", ColorConfig.TRANSITION_ACTION_COLOR, 'contains', match_type='regex')



from enum import Enum
from typing import Callable, Tuple, Optional

from qrogue.game.logic.actors import Controllable, Robot
from qrogue.game.world.navigation import Direction
from qrogue.util import CommonPopups, Logger
from .tiles import Invalid, TileCode, Tile, Wall, Floor
from .walk_trigger_tiles import WalkTriggerTile


class DoorOpenState(Enum):
    Open = 0
    Closed = 1
    KeyLocked = 2
    EventLocked = 3
    EntanglementLocked = 4


class DoorOneWayState(Enum):
    NoOneWay = 0
    Temporary = 1
    Permanent = 2


class DoorEntanglementState(Enum):
    NotEntangled = 0
    Undecided = 1
    Locked = 2
    Unlocked = 3


class Door(WalkTriggerTile):
    __INVALID_DIRECTION_IMG = Invalid().get_img()

    def __init__(self, direction: Direction, open_state: DoorOpenState = DoorOpenState.Closed,
                 one_way_state: DoorOneWayState = DoorOneWayState.NoOneWay, event_check: Callable[[], bool] = None,
                 entanglement_state: DoorEntanglementState = DoorEntanglementState.NotEntangled):
        super().__init__(TileCode.Door)
        self.__direction = direction
        self.__open_state = open_state
        self.__one_way_state = one_way_state
        self.__entanglement_state = entanglement_state
        self.__event_check = event_check
        self.__check_entanglement_lock: Optional[Callable[[], None]] = None

    @property
    def data(self) -> Tuple[Optional[DoorOpenState], Optional[DoorOneWayState], Optional[DoorEntanglementState]]:
        if self.__open_state == DoorOpenState.Open:
            # if the door is already opened we ignore the original values
            return DoorOpenState.Open, DoorOneWayState.NoOneWay, DoorEntanglementState.NotEntangled
        return self.__open_state, self.__one_way_state, self.__entanglement_state

    def set_entanglement(self, check_entanglement_lock_callback: Callable[[], None]):
        self.__check_entanglement_lock = check_entanglement_lock_callback
        self.__entanglement_state = DoorEntanglementState.Undecided

    def get_img(self):
        if self.is_open and self.__one_way_state is not DoorOneWayState.Permanent:
            return self._invisible

        if self.is_one_way:
            if self.__direction is Direction.North:
                return "^"
            elif self.__direction is Direction.East:
                return ">"
            elif self.__direction is Direction.South:
                return "v"
            elif self.__direction is Direction.West:
                return "<"
            else:
                return Door.__INVALID_DIRECTION_IMG
        else:
            if self.is_key_locked:
                return "+"
            elif self.__direction is Direction.East or self.__direction is Direction.West:
                return "|"
            else:
                return "-"

    def is_walkable(self, direction: Direction, controllable: Controllable) -> bool:
        if self.__one_way_state is DoorOneWayState.Permanent or \
                self.__one_way_state is DoorOneWayState.Temporary and not self.is_open:
            is_correct_direction = direction is self.__direction
        else:
            # opposite direction is also fine for non-one-way doors or opened temporary one-way doors
            is_correct_direction = direction in [self.__direction, self.__direction.opposite()]
        if is_correct_direction:
            if self.is_open:
                can_walk = True
            elif self.is_key_locked:
                if controllable.key_count() > 0:
                    can_walk = True
                else:
                    CommonPopups.LockedDoor.show()
                    can_walk = False
            elif self.is_event_locked:
                if self.check_event():
                    can_walk = True
                else:
                    CommonPopups.EventDoor.show()
                    can_walk = False
            elif self.is_entanglement_locked:
                CommonPopups.EntangledDoor.show()
                can_walk = False
            else:
                can_walk = True
            if self.has_explanation:
                # we overwrite a common popup that may be shown
                WalkTriggerTile._show_explanation(self._explanation, overwrite=True)
            self._explicit_trigger()
            return can_walk
        else:
            CommonPopups.WrongDirectionDoor.show()
            return False

    def _on_walk(self, direction: Direction, controllable: Controllable) -> bool:
        if self.is_key_locked:
            if isinstance(controllable, Robot):
                if not controllable.use_key():
                    Logger.instance().error(f"Error! Walked on a door without having enough keys!\n#keys="
                                            f"{controllable.key_count()}, dir={direction}", show=False,
                                            from_pycui=False)
            else:
                Logger.instance().error(f"Error! Non-Robot walked through a locked door: {controllable}", show=False,
                                        from_pycui=False)
        self.__open_state = DoorOpenState.Open
        return True

    @property
    def direction(self) -> Direction:
        return self.__direction

    @property
    def is_open(self) -> bool:
        return self.__open_state is DoorOpenState.Open and self.__entanglement_state is not DoorEntanglementState.Locked

    @property
    def is_key_locked(self) -> bool:
        return self.__open_state is DoorOpenState.KeyLocked

    @property
    def is_event_locked(self) -> bool:
        return self.__open_state is DoorOpenState.EventLocked

    @property
    def is_entanglement_locked(self) -> bool:
        if self.__entanglement_state is DoorEntanglementState.NotEntangled or \
                self.__check_entanglement_lock is None:
            return False

        if self.__entanglement_state is DoorEntanglementState.Undecided:
            if self.__check_entanglement_lock():
                self.__entanglement_state = DoorEntanglementState.Locked
            else:
                self.__entanglement_state = DoorEntanglementState.Unlocked

        return self.__entanglement_state is DoorEntanglementState.Locked

    @property
    def is_one_way(self) -> bool:
        return self.__one_way_state is not DoorOneWayState.NoOneWay

    def check_event(self) -> bool:
        # don't check again if the door is already open
        if self.is_open:
            return True

        if self.__event_check is None:
            Logger.instance().error("Tried to enter event-locked door with event_check still uninitialized!",
                                    show=False, from_pycui=False)
            self.__open_state = DoorOpenState.Closed
            return True
        if self.__event_check():
            self.__open_state = DoorOpenState.Closed
            return True
        return False

    def _copy(self) -> "Tile":
        door = Door(self.__direction, self.__open_state, self.__one_way_state, self.__event_check)
        door.set_entanglement(self.__check_entanglement_lock)
        return door

    def copy_and_adapt(self, new_direction: Direction, reset_one_way: bool = False) -> "Door":
        """
        Copies a door and assigns a new direction to it. Needed to create Hallways in the text based dungeon creator.
        Since one-way doors already have their direction set normally it doesn't make sense to give them a new one.
        Hence, a copy of a one-way door will by default no longer be a one-way door.
        :param new_direction: new direction for the copied door
        :param reset_one_way: whether the door is forced to be one-way or if it should keep its one-way flag unchanged
        :return: a new door with the same attributes except for a new direction
        """
        if new_direction is None:
            Logger.instance().throw(ValueError("Tried to copy a door with new_direction being None!"))
        if reset_one_way:
            one_way_state = DoorOneWayState.NoOneWay
        else:
            one_way_state = self.__one_way_state
        door = Door(new_direction, self.__open_state, one_way_state, self.__event_check)
        door.set_explanation(self._explanation)
        door.set_event(self._event_id)
        door.set_entanglement(self.__check_entanglement_lock)
        return door


class HallwayEntrance(Tile):
    __WALL_IMG = Wall.img()
    __FLOOR_IMG = Floor.img()

    def __init__(self, door_ref: Door):
        super().__init__(TileCode.HallwayEntrance)
        self.__door_ref = door_ref

    def get_img(self):
        if self.__door_ref.is_event_locked and not self.__door_ref.check_event():
            return self.__WALL_IMG
        return self.__FLOOR_IMG

    def is_walkable(self, direction: Direction, controllable: Controllable) -> bool:
        return not self.__door_ref.is_event_locked

    def copy(self) -> "Tile":
        return HallwayEntrance(self.__door_ref)



# exporting
from .callbacks import CallbackPack
from .expedition_map import ExpeditionMap
from .level_map import LessonMap
from .map import MapType, MapMetaData, Map
from .rooms import AreaType, Area, Hallway, Room

# importing
# +logic (Message)
# +logic.actors
# +logic.collectibles
# +world.navigation
# +world.tiles
# +util
